// music - 8bit archade4 from www.dl-sounds.com

#include "pokitto.h"
#include "tetris_gfx.h"
#include "easing.h"

#define REPSPEED 12

Pokitto::Core game;

#define HELD 0
#define NEW 1
#define RELEASE 2
byte CompletePad, ExPad, TempPad, myPad;
bool _A[3], _B[3], _C[3], _Up[3], _Down[3], _Left[3], _Right[3];
bool bgNum = 1;
char musicName[] = "pokittris.raw";
byte palNum = 0;

void UPDATEPAD(int pad, int var) {
  _C[pad] = (var >> 1)&1;
  _B[pad] = (var >> 2)&1;
  _A[pad] = (var >> 3)&1;
  _Down[pad] = (var >> 4)&1;
  _Left[pad] = (var >> 5)&1;
  _Right[pad] = (var >> 6)&1;
  _Up[pad] = (var >> 7)&1;
}

void UpdatePad(int joy_code){
  ExPad = CompletePad;
  CompletePad = joy_code;
  UPDATEPAD(HELD, CompletePad); // held
  UPDATEPAD(RELEASE, (ExPad & (~CompletePad))); // released
  UPDATEPAD(NEW, (CompletePad & (~ExPad))); // newpress
}

byte updateButtons(byte var){
   var = 0;
   if (game.buttons.cBtn()) var |= (1<<1);
   if (game.buttons.bBtn()) var |= (1<<2);
   if (game.buttons.aBtn()) var |= (1<<3);
   if (game.buttons.downBtn()) var |= (1<<4);
   if (game.buttons.leftBtn()) var |= (1<<5);
   if (game.buttons.rightBtn()) var |= (1<<6);
   if (game.buttons.upBtn()) var |= (1<<7);

   return var;
}



// some globals...
long int frameNumber = 0;
long int myDelay;
long int tempTime;
int dropTime = 0;
int slideTime = 0;
byte gameMode = 0;
bool paused = 0;
byte menuItem;
bool bgmusic = 1;

byte px, py, ps, pr; // player x,y,shape,rotation
byte nextTile;
byte lineCount, level;
int score, lines;
bool okToContinue = 0;
byte slideSpeed = 10;

void drawShape(byte x1, signed char y1, byte shape, byte frame) {
  for (char y = 0; y < 4; y++) {
    if (y1 + y > 1) {
      for (char x = 0; x < 4; x++) {
        byte mt = pgm_read_byte(shapeMap + (x + 4 * y) + shape * 64 + (frame * 16));
        if (mt > 1) {
          game.display.drawBitmap((x1 + x)*8, (y1 + y)*8, tile_gfx[(mt-1)+shape*6]);
        }
      }
    }
  }
}

bool check(signed char x1, signed char y1, char rot) {
    byte ret=0;
  for (char y = 0; y < 4; y++) {
    if (y1 + y >= 0) {
      for (char x = 0; x < 4; x++) {
        byte mt = pgm_read_byte(shapeMap + (x + 4 * y) + ps * 64 + (rot * 16));
        if (mt > 1) {
          if ( playfield[(x1 + x) + 16 * (y1 + y)] != 0) {
            ret= 1;
          }
        }
      }
    }
  }
  return ret;
}

void stamp(signed char x1, signed char y1, byte shape, byte rot) {
  for (char y = 0; y < 4; y++) {
    if (y1 + y >= 0) {
      for (char x = 0; x < 4; x++) {
        byte mt = pgm_read_byte(shapeMap + (x + 4 * y) + shape * 64 + (rot * 16));
        if (mt > 1) {
          playfield[(x1 + x) + 16 * (y1 + y)] = (mt-1)+shape*6;
        }
      }
    }
  }
}


void clearPlayfield() {
  for (char y = 0; y < 19; y++) {
    for (char x = 3; x < 13; x++) {
      playfield[x + 16 * y] = 0;
    }
  }
}

void drawBackground(){
  for (char y = 0; y < 22; y++) {
    for (char x = 0; x < 28; x++) {
      byte mt = bg_map[x + 28 * y + (616*bgNum)];
      game.display.drawBitmap(x*8, y*8, bg_gfx[mt]);
    }
  }
}

void drawPlayfield() {
    drawBackground();

  for (char y = 1; y < 19; y++) {
    for (char x = 3; x < 13; x++) {
      byte mt = playfield[x + 16 * y];
      game.display.drawBitmap((x + 5)*8, (y+1 )*8, tile_gfx[mt]);
    }
  }
    // current shape
    drawShape(px+5, py+1, ps, pr);
    // next shape?
    drawShape(19, 14, nextTile, 0);

    char text[] = "        ";
    sprintf(text, "%05d", score);
    game.display.setCursor(152,56);
    game.display.color=3;
    game.display.print(text);
    sprintf(text, "%5d", level);
    game.display.setCursor(16,96);
    game.display.print(text);
    sprintf(text, "%5d", lines);
    game.display.setCursor(16,48);
    game.display.print(text);

}



void checkLine() {
  if (py <= 1) {
    gameMode = 3;  // gameOver if off top of screen
    frameNumber=0;
  }

  score++; // increase score here as it's called whenever a tile drops

  for (char y = 0; y < 19; y++) {
    char line = 0;
    for (char x = 3; x < 13; x++) {
      line += playfield[x + 16 * y] != 0 ? 1 : 0;
    }
    if (line == 10) { // remove line
      lineCount++;
      if (lineCount == 10) {
        lineCount = 0;
        level++;
      }
      lines++;
      score += 10;



      unsigned short *p;
      palNum = level & 31;
      p=pallet+(palNum*4);
      unsigned short curPal[4];
      curPal[0] = (int)p;
      curPal[1] = (int)p+1;
      curPal[2] = (int)p+2;
      curPal[3] = (int)p+3;

      game.display.load565Palette(p);

/*
    unsigned char red = (buf & 0xf800) >> 11;
    unsigned char green = (buf & 0x07e0) >> 5;
    unsigned char blue = buf & 0x001f;
*/

      for (char y1 = y; y > 0; y--) {
        for (char x = 3; x < 13; x++) {
          playfield[x + 16 * y] = playfield[x + 16 * (y - 1)];
        }
      }
    }
  }
}


// transparent 2bit bitmap with mask
void drawMyBitmap(int16_t x, int16_t y, const uint8_t* bitmap, const uint8_t* mask)
{
    int16_t w = *bitmap;
	int16_t h = *(bitmap + 1);
    bitmap = bitmap + 2; //add an offset to the pointer to start after the width and height
    /** visibility check */
    if (y<-h || y>game.display.height) return; //invisible
    if (x<-w || x>game.display.width) return;  //invisible

    /** 2 bpp mode */
    int16_t i, j, byteNum, bitNum, byteWidth = w >> 2;
    for (i = 0; i < w; i++) {
        byteNum = i / 4;
        bitNum = (i % 4)<<1;
        for (j = 0; j < h; j++) {
            uint8_t source = *(bitmap + j * byteWidth + byteNum);
            uint8_t source2 = *(mask + j * byteWidth + byteNum+2);
            uint8_t output = (source & (0xC0 >> bitNum));
            output >>= (6-bitNum);

            uint8_t output2 = (source2 & (0xC0 >> bitNum));
            output2 >>= (6-bitNum);

            if (output2 != 0) {
                game.display.setColor(output);
                game.display.drawPixel(x + i, y + j);
            }
        }

    //return;
    }
}



void titleScreen(){
  // background
  for (char y = 0; y < 22; y++) {
    for (char x = 0; x < 28; x++) {
      byte mt = bg_map[x + 28 * y];
      game.display.drawBitmap(x*8, y*8, bg_gfx[mt]);
    }
  }

    int y=48;
    if(frameNumber<=64){
                        // time, start, distance, duration
        y = easeOutBounce(frameNumber, -48, 48+48, 64);
    }
    drawMyBitmap(16, y, title_bitmap, title_mask);

    char text[] = "Press A to Start";
    game.display.setCursor(48,120);
    game.display.color=3;
    game.display.print(text);

  if(_A[NEW]){
    gameMode = 1;
  }

}

void gameOver(){
  // background
  for (char y = 0; y < 22; y++) {
    for (char x = 0; x < 28; x++) {
      byte mt = bg_map[x + 28 * y];
      game.display.drawBitmap(x*8, y*8, bg_gfx[mt]);
    }
  }

    int y=48;
    if(frameNumber<=64){
                        // time, start, distance, duration
        y = easeOutBounce(frameNumber, -48, 48+48, 64);
    }
    drawMyBitmap(16, y, title_bitmap, title_mask);

    char text[] = "Game Over";
    game.display.setCursor(48,120);
    game.display.color=3;
    game.display.print(text);

  if(_A[NEW]){
    gameMode = 0;
    frameNumber = 0;
  }

}

void playGame(){

    if (_Left[NEW]) {
      if (check(px - 1, py, pr) == 0) {
        px--;
        slideTime = 0;
      }
    }
    if (_Right[NEW]) {
      if (check(px + 1, py, pr) == 0) {
        px++;
        slideTime = 0;
      }
    }
    if (_Left[HELD] && slideTime++ > 6) {
      if (check(px - 1, py, pr) == 0) {
        px--;
        slideTime = 12;
      }
    }
    if (_Right[HELD] && slideTime++ > 6) {
      if (check(px + 1, py, pr) == 0) {
        px++;
        slideTime = 12;
      }
    }

    if ((_Down[HELD]) || (dropTime++ > 20 - (level * 2))) {
      dropTime = 0;
      if (check(px, py+1, pr) == 0) {
        py++;
      } else {
        // place shape and create new one
        stamp(px, py, ps, pr);
        checkLine();
        py = 0; px = 6; ps = nextTile; nextTile = random(6); pr = 0;
      }
    }

    if (_A[NEW]) {
      if (check(px, py, (pr - 1) & 3) == 0) {
        pr--;
      } else if (check(px - 1, py, (pr - 1) & 3) == 0) {
        pr--; px--;
      } else if (check(px + 1, py, (pr - 1) & 3) == 0) {
        pr--; px++;
      } else if (check(px - 2, py, (pr - 1) & 3) == 0) {
        pr--; px -= 2;
      } else if (check(px + 2, py, (pr - 1) & 3) == 0) {
        pr--; px += 2;
      }
      pr &= 3;
    }

    if (_B[NEW]) {
      if (check(px, py, (pr + 1) & 3) == 0) {
        pr++;
      } else if (check(px - 1, py, (pr + 1) & 3) == 0) {
        pr++; px--;
      } else if (check(px + 1, py, (pr + 1) & 3) == 0) {
        pr++; px++;
      } else if (check(px - 2, py, (pr + 1) & 3) == 0) {
        pr++; px -= 2;
      } else if (check(px + 2, py, (pr + 1) & 3) == 0) {
        pr++; px += 2;
      }
      pr &= 3;
    }


    // render screen
    drawPlayfield();
}




int main(){

	game.begin();
	game.display.width = 220; // full size
	game.display.height = 176;
    game.display.setFont(fontC64);
    game.display.charSpacingAdjust = 0; //needed for the non-proportional C64 font (normal value=1)
    game.display.fixedWidthFont = true;

    const unsigned short *p;
    p=pallet+(palNum*4);
    game.display.load565Palette(p);

    px=6;
    gameMode=0;
    myDelay=40;

    game.sound.playMusicStream(musicName);

    frameNumber=0;

   while (game.isRunning()) {


        // if it is time to update the screen
        if (game.update()){
            frameNumber++;
            game.sound.updateStream();
            game.buttons.update();
            myPad = updateButtons(myPad);
            UpdatePad(myPad);

            char oldMode = gameMode;
            switch (gameMode) {
            case 0:
              titleScreen();
                break;
            case 1:
                if (paused) {
            //        pauseMenu();
                } else {
                    playGame();
                }
                break;
            case 3:
              gameOver();
                break;
            }
        }
    }
    return 0;
}
